from . import SMBpmPoint, SMFakeObject, SMHitObject, SMHoldObject, SMKeySoundObject, SMLiftObject, SMMapObject, \
SMMapObjectMeta, SMMapSetObject, SMMapSetObjectMeta, SMMineObject, SMRollObject, SMStop



from reamber.sm.SMMapSetObject import SMMapSetObject, SMMapObject
from reamber.osu.OsuMapObject import OsuMapObject
from reamber.osu.OsuHitObject import OsuHitObject
from reamber.osu.OsuHoldObject import OsuHoldObject
from reamber.osu.OsuBpmPoint import OsuBpmPoint
from reamber.base.BpmPoint import BpmPoint
from reamber.base.NoteObject import NoteObject
from typing import List


class SMToOsu:
    @staticmethod
    def convert(sm: SMMapSetObject) -> List[OsuMapObject]:
        """ Converts a Mapset to possibly multiple osu maps
        Note that a mapset contains maps, so a list would be expected.
        SMMap conversion is not possible due to lack of SMMapset Metadata
        :param sm: The MapSet
        :return: Osu Map
        """
        osuMapSet: List[OsuMapObject] = []
        for smMap in sm.maps:
            assert isinstance(smMap, SMMapObject)
            notes: List[NoteObject] = []

            # Note Conversion
            for note in smMap.hitObjects():
                notes.append(OsuHitObject(offset=note.offset, column=note.column))
            for note in smMap.holdObjects():
                notes.append(OsuHoldObject(offset=note.offset, column=note.column, length=note.length))

            bpms: List[BpmPoint] = []

            # Timing Point Conversion
            for bpm in smMap.bpmPoints:
                bpms.append(OsuBpmPoint(offset=bpm.offset, bpm=bpm.bpm))

            # Extract Metadata
            osuMap = OsuMapObject(
                backgroundFileName=sm.background,
                title=sm.title,
                titleUnicode=sm.titleTranslit,
                artist=sm.artist,
                artistUnicode=sm.artistTranslit,
                audioFileName=sm.music,
                creator=sm.credit,
                version=f"{smMap.difficulty} {smMap.difficultyVal}",
                previewTime=int(sm.sampleStart),
                bpmPoints=bpms,
                noteObjects=notes
            )
            osuMapSet.append(osuMap)
        return osuMapSet




from reamber.sm.SMMapSetObject import SMMapSetObject, SMMapObject
from reamber.osu.OsuMapObject import OsuMapObject
from reamber.base.BpmPoint import BpmPoint
from reamber.base.NoteObject import NoteObject
from reamber.sm.SMMapObjectMeta import SMMapObjectChartTypes
from reamber.sm.SMHitObject import SMHitObject
from reamber.sm.SMHoldObject import SMHoldObject
from reamber.sm.SMBpmPoint import SMBpmPoint
from typing import List


class OsuToSM:
    @staticmethod
    def convert(osu: OsuMapObject) -> SMMapSetObject:
        """ Converts Osu to a SMMapset Object
        Note that each osu map object will create a separate mapset, they are not merged
        :param osu: The Osu Map itself
        :return: A SM MapSet
        """
        notes: List[NoteObject] = []

        for note in osu.hitObjects():
            notes.append(SMHitObject(offset=note.offset, column=note.column))
        for note in osu.holdObjects():
            notes.append(SMHoldObject(offset=note.offset, column=note.column, length=note.length))

        bpms: List[BpmPoint] = []

        for bpm in osu.bpmPoints:
            bpms.append(SMBpmPoint(offset=bpm.offset, bpm=bpm.bpm))

        smSet: SMMapSetObject = SMMapSetObject(
            music=osu.audioFileName,
            title=osu.title,
            titleTranslit=osu.titleUnicode,
            artist=osu.artist,
            artistTranslit=osu.artistUnicode,
            credit=osu.creator,
            background=osu.backgroundFileName,
            sampleStart=osu.previewTime,
            sampleLength=10,
            offset=osu.firstNoteOffset(),
            maps=[
                SMMapObject(
                    chartType=SMMapObjectChartTypes.DANCE_SINGLE,
                    noteObjects=notes,
                    bpmPoints=bpms
                )
            ]
        )

        return smSet






from __future__ import annotations
from reamber.osu.OsuTimingPointMeta import OsuTimingPointMeta
from reamber.base.TimedObject import TimedObject
from dataclasses import dataclass


@dataclass
class OsuSliderVelocity(OsuTimingPointMeta, TimedObject):
    velocity: float = 1.0

    @staticmethod
    def codeToValue(code: float) -> float:
        return -100.0 / code

    @staticmethod
    def valueToCode(value: float) -> float:
        return -100.0 / value

    @staticmethod
    def readString(s: str) -> OsuSliderVelocity:
        if s.isspace():
            return None

        sComma = s.split(",")
        if len(sComma) < 8:
            return None

        this = OsuSliderVelocity()
        this.offset = float(sComma[0])
        this.velocity = OsuSliderVelocity.codeToValue(float(sComma[1]))
        this.sampleSet = int(sComma[3])
        this.sampleSetIndex = int(sComma[4])
        this.volume = int(sComma[5])
        this.kiai = int(sComma[7])

        return this

    def writeString(self) -> str:
        return f"{int(self.offset)},{self.valueToCode(self.velocity)}," \
               f"4,{self.sampleSet}," \
               f"{self.sampleSetIndex},{self.volume},{0},{int(self.kiai)}"



from typing import List
from dataclasses import dataclass
from dataclasses import field

from reamber.base.MapObject import MapObject


@dataclass
class MapSetObject:
    maps: List[MapObject] = field(default_factory=lambda: [])



from reamber.osu.OsuSampleSet import OsuSampleSet
from dataclasses import dataclass
from math import ceil


class OsuHitObjectType:
    NOTE: int = 1
    LONG_NOTE: int = 128


@dataclass
class OsuHitObjectMeta:
    hitsoundSet: int = OsuSampleSet.AUTO
    sampleSet: int = OsuSampleSet.AUTO
    additionSet: int = OsuSampleSet.AUTO
    customSet: int = OsuSampleSet.AUTO
    volume: int = 0
    hitsoundFile: str = ""
    # keys to be supplied by map

    @staticmethod
    def xAxisToColumn(xAxis: float, keys: int) -> int:
        return int(ceil((xAxis * keys - 256.0) / 512.0))

    @staticmethod
    def columnToXAxis(column: float, keys: int) -> int:
        return int(round(((512.0 * column) + 256.0) / keys))

    @staticmethod
    def isHitObject(s: str):
        return s.count(":") == 4

    @staticmethod
    def isHoldObject(s: str):
        return s.count(":") == 5



from typing import List
from typing import Tuple
from dataclasses import dataclass
from dataclasses import field

from reamber.base.NoteObject import NoteObject
from reamber.base.HitObject import HitObject
from reamber.base.HoldObject import HoldObject
from reamber.base.BpmPoint import BpmPoint


@dataclass
class MapObject:
    noteObjects: List[NoteObject] = field(default_factory=lambda: [])
    bpmPoints: List[BpmPoint] = field(default_factory=lambda: [])

    def noteObjectsSorted(self) -> List[NoteObject]:
        """ Returns a copy of Sorted NoteObjs """
        return sorted(self.noteObjects, key=lambda ho: ho.offset)

    def bpmPointsSorted(self) -> List[BpmPoint]:
        """ Returns a copy of Sorted BPMs """
        return sorted(self.bpmPoints, key=lambda tp: tp.offset)

    def hitObjects(self, sort: bool = False) -> List[HitObject]:
        """ Returns a copy of (Sorted) HitObjs """
        if sort:
            return sorted([note for note in self.noteObjects if isinstance(note, HitObject)], key=lambda x: x.offset)
        else:
            return [note for note in self.noteObjects if isinstance(note, HitObject)]

    def holdObjects(self, sort: bool = False) -> List[HoldObject]:
        """ Returns a copy of (Sorted) HoldObjs """
        if sort:
            return sorted([note for note in self.noteObjects if isinstance(note, HoldObject)], key=lambda x: x.offset)
        else:
            return [note for note in self.noteObjects if isinstance(note, HoldObject)]

    def hitObjectOffsets(self, sort: bool = True) -> List[float]:
        """ Returns a copy of the HitObj Offsets """
        return [ho.offset for ho in self.hitObjects(sort)]

    def holdObjectOffsets(self, sort: bool = True) -> List[Tuple[float, float]]:
        """ Returns a copy of the HoldObj Offsets [(Head0, Tail0), (Head1, Tail1), ...]"""
        return [(ho.offset, ho.tailOffset()) for ho in self.holdObjects(sort)]

    def addNoteOffsets(self, by: float):
        """ Move all notes by a specific ms """
        for note in self.noteObjects: note.offset += by

    def addBpmOffsets(self, by: float):
        """ Move all bpms by a specific ms """
        for bpm in self.bpmPoints: bpm.offset += by

    def addOffset(self, by: float):
        """ Move all by a specific ms """
        self.addNoteOffsets(by)
        self.addBpmOffsets(by)

    def lastNoteOffset(self) -> float:
        """ Get Final Note Offset """
        hos = self.noteObjectsSorted()
        lastHit = self.hitObjects()[-1]
        lastHold = self.holdObjects()[-1]

        if lastHit.offset > lastHold.offset + lastHold.length:
            return lastHit.offset
        else:
            return lastHold.offset + lastHold.length

    def firstNoteOffset(self) -> float:
        """ Get First Note Offset """
        hos = self.noteObjectsSorted()
        return hos[0].offset



from reamber.osu.OsuMapObjectMeta import OsuMapObjectMeta
from reamber.base.MapObject import MapObject

from reamber.osu.OsuTimingPointMeta import OsuTimingPointMeta
from reamber.osu.OsuBpmPoint import OsuBpmPoint
from reamber.osu.OsuSliderVelocity import OsuSliderVelocity

from reamber.osu.OsuHitObject import OsuHitObject
from reamber.osu.OsuHoldObject import OsuHoldObject
from reamber.osu.OsuHitObjectMeta import OsuHitObjectMeta

from typing import List
from dataclasses import dataclass
from dataclasses import field


@dataclass
class OsuMapObject(MapObject, OsuMapObjectMeta):

    svPoints: List[OsuSliderVelocity] = field(default_factory=lambda: [])

    def readFile(self, filePath=""):
        with open(filePath, "r", encoding="utf8") as f:
            file = f.read()
            file = file.replace("[TimingPoints]\n", "[HitObjects]\n")  # This is so as to split multiple delimiters
            fileSpl = file.split("[HitObjects]\n")
            if len(fileSpl) != 3:
                return

            self._readFileMetadata(fileSpl[0].split("\n"))

            for line in fileSpl[1].split("\n"):
                self._readFileTimingPoints(line)

            for line in fileSpl[2].split("\n"):
                self._readFileHitObjects(line)

    def writeFile(self, filePath=""):
        with open(filePath, "w+", encoding="utf8") as f:
            for s in self.writeStringList():
                f.write(s + "\n")

            f.write("\n[TimingPoints]\n")
            for tp in self.bpmPoints:
                assert isinstance(tp, OsuBpmPoint)
                f.write(tp.writeString() + "\n")

            for tp in self.svPoints:
                assert isinstance(tp, OsuSliderVelocity)
                f.write(tp.writeString() + "\n")

            f.write("\n[HitObjects]\n")
            for ho in self.noteObjects:
                assert isinstance(ho, (OsuHitObject, OsuHoldObject))
                f.write(ho.writeString(keys=int(self.circleSize)) + "\n")

    def _readFileMetadata(self, lines: List[str]):
        self.readStringList(lines)

    def _readFileTimingPoints(self, line: str):
        if OsuTimingPointMeta.isSliderVelocity(line):
            self.svPoints.append(OsuSliderVelocity.readString(line))
        elif OsuTimingPointMeta.isTimingPoint(line):
            self.bpmPoints.append(OsuBpmPoint.readString(line))

    def _readFileHitObjects(self, line: str):
        if OsuHitObjectMeta.isHitObject(line):
            self.noteObjects.append(OsuHitObject.readString(line, int(self.circleSize)))
        elif OsuHitObjectMeta.isHoldObject(line):
            self.noteObjects.append(OsuHoldObject.readString(line, int(self.circleSize)))




from __future__ import annotations
from reamber.osu.OsuTimingPointMeta import OsuTimingPointMeta
from reamber.base.BpmPoint import BpmPoint
from dataclasses import dataclass


@dataclass
class OsuBpmPoint(OsuTimingPointMeta, BpmPoint):
    @staticmethod
    def codeToValue(code: float) -> float:
        return 60000.0 / code

    @staticmethod
    def valueToCode(value: float) -> float:
        return 60000.0 / value

    @staticmethod
    def readString(s: str) -> OsuBpmPoint:
        if s.isspace(): return None

        sComma = s.split(",")
        if len(sComma) < 8: return None

        this = OsuBpmPoint()
        this.offset = float(sComma[0])
        this.bpm = OsuBpmPoint.codeToValue(float(sComma[1]))
        this.metronome = int(sComma[2])
        this.sampleSet = int(sComma[3])
        this.sampleSetIndex = int(sComma[4])
        this.volume = int(sComma[5])
        this.kiai = int(sComma[7])

        return this

    def writeString(self) -> str:
        return f"{int(self.offset)},{self.valueToCode(self.bpm)}," \
               f"{self.metronome},{self.sampleSet}," \
               f"{self.sampleSetIndex},{self.volume},{1},{int(self.kiai)}"



from reamber.base.MapSetObject import MapSetObject
from reamber.sm.SMMapSetObjectMeta import SMMapSetObjectMeta
from reamber.sm.SMMapObject import SMMapObject
from reamber.sm.SMStop import SMStop
from dataclasses import dataclass
from reamber.sm.SMBpmPoint import SMBpmPoint

from reamber.base.RAConst import RAConst

from typing import List


@dataclass
class SMMapSetObject(MapSetObject, SMMapSetObjectMeta):

    def readFile(self, filePath: str):
        with open(filePath, "r", encoding="utf8") as f:
            file = f.read()
            fileSpl = file.split(";")
            metadata = []
            maps = []
            for index, line in enumerate(fileSpl):
                try:
                    line.index("#NOTES:")
                    maps .append(line)
                except ValueError:
                    metadata.append(line)

            self._readMetadata(metadata)
            bpms = self._readBpms(offset=self.offset, lines=self._bpmsStr)
            self._readStops(lines=self._stopsStr, bpms=bpms)
            self._readMaps(maps=maps, bpms=bpms, stops=self.stops)

            for map in self.maps:
                map.bpmPoints = bpms
            return

    def writeFile(self, filePath: str,
                  alignBpms: bool = False,
                  BEAT_CORRECTION_FACTOR = 5.0,
                  BEAT_ERROR_THRESHOLD = 0.001):
        """
        Writes the file to filePath specified
        :param BEAT_ERROR_THRESHOLD: See BpmPoint.py::alignBpms for details
        :param BEAT_CORRECTION_FACTOR: See BpmPoint.py::alignBpms for details
        :param filePath: File Path
        :param alignBpms: Aligns the BPM by mutating the current file. Details in BpmPoint.py
        """
        with open(filePath, "w+", encoding="utf8") as f:
            if alignBpms:
                for map in self.maps:
                    map.bpmPoints = SMBpmPoint.alignBpms(map.bpmPoints,
                                                         BEAT_CORRECTION_FACTOR=BEAT_CORRECTION_FACTOR,
                                                         BEAT_ERROR_THRESHOLD=BEAT_ERROR_THRESHOLD)
            for s in self._writeMetadata(self.maps[0].bpmPoints):
                f.write(s + "\n")

            for map in self.maps:
                assert isinstance(map, SMMapObject)
                for s in map.writeString():
                    f.write(s + "\n")

    @staticmethod
    def _readBpms(offset: float, lines: List[str]) -> List[SMBpmPoint]:
        assert offset is not None, "Offset should be defined BEFORE Bpm"
        bpms = []
        beatPrev = 0.0
        bpmPrev = 1.0
        for line in lines:
            beatCurr, bpmCurr = [float(x.strip()) for x in line.split("=")]
            offset += (beatCurr - beatPrev) * RAConst.minToMSec(1.0 / bpmPrev)
            bpms.append(SMBpmPoint(offset=offset, bpm=bpmCurr))
            beatPrev = beatCurr
            bpmPrev = bpmCurr

        return bpms

    def _readStops(self, bpms: List[SMBpmPoint], lines: List[str]):
        for line in lines:
            if len(line) == 0: return
            beatCurr, lengthCurr = [float(x.strip()) for x in line.split("=")]

            index = 0
            for index, bpm in enumerate(bpms):
                if bpm.beat(bpms) > beatCurr:
                    index -= 1
                    break

            offset = bpms[index].offset + (beatCurr - bpms[index].beat(bpms)) * bpms[index].beatLength()

            self.stops.append(SMStop(offset=offset, length=RAConst.secToMSec(lengthCurr)))

    def _readMaps(self, maps: List[str], bpms: List[SMBpmPoint], stops: List[SMStop]):
        for map in maps:
            self.maps.append(SMMapObject.readString(map=map, bpms=bpms, stops=stops))



from reamber.base.NoteObject import NoteObject
from dataclasses import dataclass


@dataclass
class HoldObject(NoteObject):
    length: float = 0.0

    def tailOffset(self) -> float:
        """ Gets the offset for the tail """
        return self.offset + self.length



class RAConst :

    HR_TO_MIN   : float = 60.0
    HR_TO_SEC   : float = 60.0 * 60.0
    HR_TO_MSEC  : float = 60.0 * 60.0 * 1000.0
    MIN_TO_HR   : float = 1.0 / 60.0
    MIN_TO_SEC  : float = 60.0
    MIN_TO_MSEC : float = 60.0 * 1000.0
    SEC_TO_HR   : float = 1.0 / (60.0 * 60.0)
    SEC_TO_MIN  : float = 1.0 / 60.0
    SEC_TO_MSEC : float = 1000.0
    MSEC_TO_HR  : float = 1.0 / (60.0 * 60.0 * 1000.0)
    MSEC_TO_MIN : float = 1.0 / (60.0 * 1000.0)
    MSEC_TO_SEC : float = 1.0 / 1000.0

    @staticmethod
    def hrToMin(hours): return float(hours * RAConst.HR_TO_MIN)
    @staticmethod
    def hrToSec(hours): return float(hours * RAConst.HR_TO_SEC)
    @staticmethod
    def hrToMSec(hours): return float(hours * RAConst.HR_TO_MSEC)
    @staticmethod
    def minToHr(mins): return float(mins * RAConst.MIN_TO_HR)
    @staticmethod
    def minToSec(mins): return float(mins * RAConst.MIN_TO_SEC)
    @staticmethod
    def minToMSec(mins): return float(mins * RAConst.MIN_TO_MSEC)
    @staticmethod
    def secToHr(secs): return float(secs * RAConst.SEC_TO_HR)
    @staticmethod
    def secToMin(secs): return float(secs * RAConst.SEC_TO_MIN)
    @staticmethod
    def secToMSec(secs): return float(secs * RAConst.SEC_TO_MSEC)
    @staticmethod
    def mSecToHr(msecs): return float(msecs * RAConst.MSEC_TO_HR)
    @staticmethod
    def mSecToMin(msecs): return float(msecs * RAConst.MSEC_TO_MIN)
    @staticmethod
    def mSecToSec(msecs): return float(msecs * RAConst.MSEC_TO_SEC)
